/*--

   Copyright (c) 2004 Ambros Marzetta and David Bremner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

--*/
/* z_bt1.c (ZRAM: Backtracking (sequential depth first)
   Ambros Marzetta

 * $Source$
 * $Revision: 5167 $
 * $Date: 2004-10-20 13:20:33 -0300 (Wed, 20 Oct 2004) $
 * $Log$
 * Revision 1.3  2004/10/20 16:20:32  bremner
 * added copyright notice and GPL boilerplate
 *
 * Revision 1.2  2000/11/15 16:14:52  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.1  1998/03/10  15:04:39  marzetta
 * Initial revision
 *
 */

#include "z_bt.h"
#include "z_lb.h"
#include "z_mp.h"
#include <stdlib.h>

static boolean bt1_inited;

/* All procedures stop and return the current node when the branching
   procedure returns true. They return NULL if the branching procedure
   always returns false.
   */

/* Variables ----------------------------------------------------------- */

static struct {
  z_bt_GlobalP app_global;
  z_bt_BranchProc BranchInPlace;
  s_seqP workList;
  z_NodeP solution;
} glob;

/* Sequential depth first ---------------------------------------------- */

static struct {
  z_NodeP solution;
  z_bt_ProcSetP psp;
  z_NodeP global;
} dfs_glob;

static void bt_do_df_child(z_NodeP child, void *glob)
{
  boolean stop;

  if (! dfs_glob.solution) {
    stop = dfs_glob.psp->bt_BranchInPlace(dfs_glob.global,
					  child, bt_do_df_child, 0);
    if (stop) dfs_glob.solution = z_NewCopyNode(child);
  }
} /* bt_do_df_child */

z_NodeP z_bt_EnumerateDepthFirst(z_ProcSetNum procsetnum,
				 z_bt_GlobalP global,
				 z_NodeP root)
{
  /* Enumerate all nodes starting from root (incl).
   */
  boolean stop;
  
  assert(bt1_inited);
  assert(root);
  dfs_glob.global = global;
  dfs_glob.psp = (z_bt_ProcSetP) z_GetProcSet(procsetnum);
  assert(dfs_glob.psp->bt_BranchInPlace);
  dfs_glob.solution = NULL;
  stop = dfs_glob.psp->bt_BranchInPlace(global, root, bt_do_df_child, 0);
  if (stop) dfs_glob.solution = z_NewCopyNode(root);
  return dfs_glob.solution;
} /* z_bt_EnumerateDepthFirst */


/* Parallel / Sequential using z_lb.h ------------------------------------ */

static z_ProcSetNum bt_procsetnum;
static z_lb_ProcSet bt_procset;
static int bt_GlobalTag;

typedef struct bt_Global {
  Z_NODE_HEADER;		/* len; tag = bt_GlobalTag */
  z_NodeP bt_app_global;	/* application global node */
  int bt_app_psn;		/* application procedure set (z_bt_ProcSetP) */
} bt_Global, *bt_GlobalP;

static bt_GlobalP NewBtGlobal(z_ProcSetNum procsetnum, z_lb_GlobalP ap_global)
{
  bt_GlobalP bt_global;

  assert(ap_global);
  bt_global = (bt_GlobalP) z_NewNode(bt_GlobalTag, sizeof(bt_Global));
  bt_global->bt_app_psn = procsetnum;
  bt_global->bt_app_global = ap_global;
  return bt_global;
} /* NewRsGlobal */

void z_bt_lb_Do(z_NodeP np, void *ref)
{
  /* This short procedure must be as efficient as possible*/

  if (!z_lb_needSplit) {	/* depth first */
#ifndef SEQUENTIAL
    if (z_lb_needPoll)
      z_Poll();			/* this call doubles the running time */
				/* on tobi for set partitions */
#endif /* SEQUENTIAL */
    if (glob.BranchInPlace(glob.app_global, np, z_bt_lb_Do, ref)) {
      assert(0);
    }
  } else			/* insert */
    (void) s_seq_InsertFirst(glob.workList, z_NewCopyNode(np));
} /* z_bt_lb_Do */

void bt_WorkAndSplit(bt_GlobalP global,
		     z_lb_WorkUnitP workUnit,
		     s_seqP workList)
     /* called from z_lb through bt_procset */
{
  z_bt_ProcSetP psp;

  psp = (z_bt_ProcSetP) z_GetProcSet(global->bt_app_psn);
  glob.BranchInPlace = psp->bt_BranchInPlace;
  glob.workList = workList;
  glob.app_global = global->bt_app_global;
  glob.solution = NULL;
  glob.BranchInPlace(glob.app_global, workUnit, &z_bt_lb_Do, NULL);
} /* bt_WorkAndSplit */

z_NodeP z_bt_EnumerateLB(z_ProcSetNum procsetnum,
			 z_bt_GlobalP global,
			 z_NodeP root,
			 boolean sequential)
{
  bt_GlobalP bt_global;

  bt_global = NewBtGlobal(procsetnum, global);
  z_lb_Work(bt_procsetnum, bt_global, root, sequential);
  return NULL;
} /* z_bt_EnumerateLB */


/* Sequential or Parallel ----------------------------------------------- */

z_NodeP z_bt_Enumerate(z_ProcSetNum procsetnum,
		       z_bt_GlobalP global,
		       z_NodeP root,
		       boolean sequential)
{
  return z_bt_EnumerateLB(procsetnum, global, root, sequential);
} /* z_bt_Enumerate */


/* Initialisation ------------------------------------------------------- */

void z_bt1_Init(void)
{
  /* Initialize this module */
  if (! bt1_inited) {
    bt1_inited = 1;
    bt_GlobalTag = z_InstallClass(bt_Global, 1, 0, 0);
    z_lb_Init();
    bt_procset.lb_Work = NULL;
    bt_procset.lb_Split = NULL;
    bt_procset.lb_WorkAndSplit = (z_lb_SplitProc) bt_WorkAndSplit;
    bt_procset.lb_Fini = NULL;
    bt_procsetnum = z_InstallProcSet((z_ProcSetP) &bt_procset);
  }
} /* z_bt1_Init */

